#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "as_tree.h"
#include "machine_generic.h"
#include "quad.h"

/*
 * We set quads handlers and run code generation. Code generation routine calls
 * generic quads handlers, each handler calls special handler.
 * We try to make the code generation procedure the same for all kind of architectures.
 *
 */

typedef void (*generic_handler_t)(struct quad *, struct quads_ctx *);

static void generic_label(struct quad *q, struct quads_ctx *ctx);
static void generic_goto(struct quad *q, struct quads_ctx *ctx);
static void generic_if(struct quad *q, struct quads_ctx *ctx);
static void generic_call(struct quad *q, struct quads_ctx *ctx);
static void generic_param(struct quad *q, struct quads_ctx *ctx);
static void generic_op(struct quad *q, struct quads_ctx *ctx);
static void generic_assign(struct quad *q, struct quads_ctx *ctx);
static void generic_store_access(struct quad *q, struct quads_ctx *ctx);
static void generic_load_access(struct quad *q, struct quads_ctx *ctx);
#ifdef _MEM_AREA
static void generic_store_mem_area(struct quad *q, struct quads_ctx *ctx);
static void generic_load_mem_area(struct quad *q, struct quads_ctx *ctx);
#endif
static void generic_func(struct quad *q, struct quads_ctx *ctx);
static void generic_scope(struct quad *q, struct quads_ctx *ctx);
static void generic_scope_end(struct quad *q, struct quads_ctx *ctx);
static void generic_func_end(struct quad *q, struct quads_ctx *ctx);

/* Generic handlers for special type of quad. */
static struct {
	quadruple_type_t quad_type;
	generic_handler_t handler;
} generics[] = {
	{ QUAD_LABEL,		generic_label },
	{ QUAD_GOTO,		generic_goto },
	{ QUAD_IF,		generic_if },
	{ QUAD_CALL,		generic_call },
	{ QUAD_PARAM,		generic_param },
	{ QUAD_OP,		generic_op },
	{ QUAD_ASSIGN,		generic_assign },
	{ QUAD_STORE_ACCESS,	generic_store_access },
	{ QUAD_LOAD_ACCESS,	generic_load_access },
#ifdef _MEM_AREA
	{ QUAD_STORE_MEM_AREA,	generic_store_mem_area },
	{ QUAD_LOAD_MEM_AREA,	generic_load_mem_area },
#endif
	{ QUAD_FUNC, 		generic_func },
	{ QUAD_SCOPE,		generic_scope },
	{ QUAD_SCOPE_END,	generic_scope_end },
	{ QUAD_FUNC_END,	generic_func_end },
	{ QUAD_UNKNOWN,		NULL }
};

/* Pointers to real handlers. */
static struct q_vfuncs gen_code_funcs;
/* Function to print architecture instructions. */
code_print_out_t code_print_out;

struct q_vfuncs *machine_get_vfuncs()
{
	return &gen_code_funcs;
}

void machine_set_vfuncs(struct q_vfuncs *funcs)
{
	return_if_fail(funcs != NULL);

	gen_code_funcs.q_assign = funcs->q_assign;
	gen_code_funcs.q_call = funcs->q_call;
	gen_code_funcs.q_func = funcs->q_func;
	gen_code_funcs.q_func_end = funcs->q_func_end;
	gen_code_funcs.q_goto = funcs->q_goto;
	gen_code_funcs.q_if = funcs->q_if;
	gen_code_funcs.q_label = funcs->q_label;
	gen_code_funcs.q_load_access = funcs->q_load_access;
	gen_code_funcs.q_store_access = funcs->q_store_access;
#ifdef _MEM_AREA
	gen_code_funcs.q_store_mem_area = funcs->q_store_mem_area;
	gen_code_funcs.q_load_mem_area = funcs->q_load_mem_area;
#endif
	gen_code_funcs.q_op = funcs->q_op;
	gen_code_funcs.q_param = funcs->q_param;
	gen_code_funcs.q_scope = funcs->q_scope;
	gen_code_funcs.q_scope_end = funcs->q_scope_end;
}

/* Default print out function. */
static void code_print_out_default(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
}

/* Below you see how we call a generic handler for a quad. */
static void generic_label(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_label(q, ctx);
}

static void generic_goto(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_goto(q, ctx);
}

static void generic_if(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_if(q, ctx);
}

static void generic_call(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_call(q, ctx);
}

static void generic_param(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_param(q, ctx);
}

static void generic_op(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_op(q, ctx);
}

static void generic_assign(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_assign(q, ctx);
}

static void generic_store_access(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_store_access(q, ctx);
}

#ifdef _MEM_AREA
static void generic_store_mem_area(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_store_mem_area(q, ctx);
}

static void generic_load_mem_area(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_load_mem_area(q, ctx);
}
#endif

static void generic_load_access(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_load_access(q, ctx);
}

static void generic_func(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_func(q, ctx);
}

static void generic_scope(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_scope(q, ctx);
}

static void generic_scope_end(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_scope_end(q, ctx);
}

static void generic_func_end(struct quad *q, struct quads_ctx *ctx)
{
	return_if_fail(q != NULL);

	gen_code_funcs.q_func_end(q, ctx);
}

/* Generate a machine dependent code from the quad's array. */
void machine_gen_code(struct quad **quads, unsigned int quads_size, code_print_out_t func)
{
	struct quads_ctx ctx;
	unsigned int i, j;

	return_if_fail(quads != NULL);

	if (func != NULL)
		code_print_out = func;
	else
		code_print_out = code_print_out_default;

	ctx.quads = quads;
	ctx.size = quads_size;

	for (i = 0; i < quads_size; i++) {
		ctx.cur = i;
		for (j = 0; generics[j].quad_type != QUAD_UNKNOWN; j++) {
			if (generics[j].quad_type == quads[i]->quad_type) {
				generics[j].handler(quads[i], &ctx);
				break;
			}
		}
	}
}
